function [A B C outmatrices cost] = NNCMTF(X,matrices,F)
%Vagelis Papalexakis, School of Computer Science, Carnegie Mellon
%University, 2013-2014
%ALS algorithm for Non-Negative Coupled Matrix-Tensor Factorization
%Example: [A B C outmatrices] = CMTF(X,matrices,F)
%   X: input tensor
%   matrices{1}: coupled matrix on 1st dimension
%   matrices{2}: coupled matrix on 2nd dimension
%   matrices{3}: coupled matrix on 3rd dimension
%   F: number of components
%   [A B C]: PARAFAC decomposition of X
%   outmatrices{i}: i-th latent factor of coupled matrix sharing i-th
%   dimension


MAXNUMITER = 100; 
SMALLNUMBER = 10^-6;
s = size(X); I=s(1);J=s(2);K=s(3);

X1 = matrices{1}; X2 = matrices{2}; X3 = matrices{3};
if size(X1,1)~=I && ~isempty(X1)
   error('X1 must have I rows'); 
end
if size(X2,1)~=J && ~isempty(X2)
   error('X2 must have J rows'); 
end
if size(X3,1)~=K && ~isempty(X3)
   error('X3 must have K rows'); 
end

D = sprand(size(X1,2),F,1); E = sprand(size(X2,2),F,1); G = sprand(size(X3,2),F,1);

UA = double(tenmat(X,1,'bc')); UA = UA';
UB = double(tenmat(X,2,'bc')); UB = UB';
UC = double(tenmat(X,3,'bc')); UC = UC';
       
OPTS.printitn = 0;factors = cp_nmu(X,F,OPTS);A = factors.U{1}; B = factors.U{2}; C = factors.U{3};%Plain PARAFAC initialization


cost = 0;
if ~isempty(X1)
    cost = cost + norm(X1 - A*D','fro')^2;
end
if ~isempty(X2)
    cost = cost + norm(X2 - B*E','fro')^2;
end
if ~isempty(X3)
    cost = cost + norm(X3 - C*G','fro')^2;
end

cost = cost + norm(UA - khatrirao(B,C)*A' ,'fro')^2;
costold = 2*cost;
it = 0;
while abs((cost-costold)/costold) > SMALLNUMBER && it < MAXNUMITER && cost > 10^5*eps
        it = it+1;
        
        A = max(0,([(khatrirao(B,C)) ; D]\[UA ; X1'])');
        % re-estimate B:
        B = max(0, ( [(khatrirao(C,A)) ; E]\[UB ; X2'])');
        % re-estimate C:
        C = max(0, ( [(khatrirao(A,B)); G ]\[UC ; X3'])');    
        
        
        costold = cost;
        
        cost = 0;
        if ~isempty(X1)
            D = max(0, ((A\X1)'));
            cost = cost + norm(X1 - A*D','fro')^2;
        else
            D = [];
        end
        if ~isempty(X2)
            E = max(0, ((B\X2)'));
            cost = cost + norm(X2 - B*E','fro')^2;
        else
            E = [];
        end
        if ~isempty(X3)
            G = max(0, ((C\X3)'));
            cost = cost + norm(X3 - C*G','fro')^2;
        else
            G = [];
        end
        cost = cost + norm(UA - khatrirao(B,C)*A' ,'fro')^2;
        fprintf('iteration: %d cost: %12.10f diff: %.12f\n',it,cost,abs((cost-costold)/costold));

end
outmatrices{1} = (D); outmatrices{2} = (E); outmatrices{3} = (G);
end