function [A B C outmatrices lambda sigma allcost] = coupled_parCube(X,matrices,F,sample_factor,times,nonneg,p)
%Vagelis Papalexakis, 2013-2014
%School of Computer Science, Carnegie Mellon University
%Implementation of the multi-core version of Turbo-SMT Algorithm using the
%CMTF toolbox as core algorithm.


if nargin == 5
    nonneg = 0;
    p = 0.8;
end
if nargin == 6
    p=0.8;
end
mypath = pwd;

lambda = zeros(6, F, times);%hold scaling factor for each column of all matrices, for each repetition

s = size(X); I = s(1); J = s(2); K = s(3);

%handle the additional matrices
mat = {};
for i = 1:length(matrices)
   mat{i} = matrices{i}; 
end
for i = length(matrices)+1 : 3
    mat{i} = [];
end
matrices = mat;

X1 = matrices{1}; X2 = matrices{2}; X3 = matrices{3};

if ~isempty(X1)
    if size(X1,1)~=I
       error('X1 must have I rows'); 
    end
end

if ~isempty(X2)
    if size(X2,1)~=J
       error('X2 must have J rows'); 
    end
end

if ~isempty(X3)
    if size(X3,1)~=K
       error('X3 must have K rows'); 
    end
end
% tic
allcost = zeros(times,1);
UC = (sptenmat(X,3,'bc'));  UC = sparse(UC.subs(:,1), UC.subs(:,2),UC.vals,size(UC,1),size(UC,2));UC = UC';
% serialtime = serialtime + toc;

A = sparse(I,F); B = sparse(J,F);C = sparse(K,F);

for i = 1:times
    allCorrespondences = zeros(F,1);
% %     tic %start timing parallelizable portion
    if i == 1
        [Xs X1s X2s X3s idx_i idx_j idx_k idx_d idx_e idx_g Xma Xmb Xmc] = coupled_parCube_core(X,X1,X2,X3,sample_factor);
        
        fixed_i = idx_i(1:ceil(p*length(idx_i)));
        fixed_j = idx_j(1:ceil(p*length(idx_j)));
        fixed_k = idx_k(1:ceil(p*length(idx_k)));
        fixed_d = idx_d(1:ceil(p*length(idx_d)));
        fixed_e = idx_e(1:ceil(p*length(idx_e)));
        fixed_g = idx_g(1:ceil(p*length(idx_g)));

        fixed{1} = fixed_i; fixed{2} = fixed_j; fixed{3}=fixed_k;
        fixed{4} = fixed_d; fixed{5} = fixed_e; fixed{6} = fixed_g;
    else 
        [Xs X1s X2s X3s idx_i idx_j idx_k idx_d idx_e idx_g Xma Xmb Xmc] = coupled_parCube_core(X,X1,X2,X3,sample_factor,fixed);     
    end
    
    mats{1} = []; mats{2} = []; mats{3} = [];
    if ~isempty(X1)
        mats{1} = X1s;
    end
    if ~isempty(X2)
        mats{2} = X2s;
    end
    if ~isempty(X3)
        mats{3} = X3s;
    end
    size(Xs)
    if nnz(Xs)>0
        Fac = cmtf_opt_wrapper(tensor(Xs),tensor(full(X1s)),F);
        As = Fac.U{1}; Bs = Fac.U{2}; Cs = Fac.U{3}; OUTMs{1} = Fac.U{4};OUTMs{2} = []; OUTMs{3} = [];
       % As=sparse(factors.U{1});Bs=sparse(factors.U{2});Cs=sparse(factors.U{3});%lambda = factors.lambda;As = As*diag(lambda);
%         As = As*diag(lambda(:,i).^(1/3)); Bs = Bs*diag(lambda(:,i).^(1/3)); Cs = Cs*diag(lambda(:,i).^(1/3));  
    else
       As = 0; Bs = 0; Cs = 0; OUTMs = {};
    end
    
    

    Atemp = sparse(I,F); Btemp = sparse(J,F);Ctemp = sparse(K,F);


    Atemp(idx_i,:) = As; Btemp(idx_j,:) = Bs; Ctemp(idx_k,:) = Cs;
    
    
    if ~isempty(X1)
        Dtemp = sparse(size(X1,2),F); 
        Dtemp(idx_d,:) = OUTMs{1};
    end
    if ~isempty(X2)
        Etemp = sparse(size(X2,2),F);
        Etemp(idx_e,:) = OUTMs{2};
    end
    if ~isempty(X3)
        Gtemp = sparse(size(X3,2),F);    
        Gtemp(idx_g,:) = OUTMs{3};
    end
   
    
     %now, normalize using the scaling of the common part of every column
    for f = 1:F
       norm_a = norm(Atemp(fixed_i,f),2);Atemp(:,f) = Atemp(:,f)/(norm_a + eps); [value,idx] = max(Atemp(fixed_i,f));Atemp(:,f) = Atemp(:,f)*sign(value); lambda(1,f,i) = sign(value)*(norm_a + eps);
       norm_b = norm(Btemp(fixed_j,f),2);Btemp(:,f) = Btemp(:,f)/(norm_b + eps); [value,idx] = max(Btemp(fixed_j,f));Btemp(:,f) = Btemp(:,f)*sign(value); lambda(2,f,i) = sign(value)*(norm_b + eps);
       norm_c = norm(Ctemp(fixed_k,f),2);Ctemp(:,f) = Ctemp(:,f)/(norm_c + eps); [value,idx] = max(Ctemp(fixed_k,f));Ctemp(:,f) = Ctemp(:,f)*sign(value); lambda(3,f,i) = sign(value)*(norm_c + eps);
        
       %consider cases for all possible numbers of out matrices
        if ~isempty(X1)
           norm_d = norm(Dtemp(fixed_d,f),2);Dtemp(:,f) = Dtemp(:,f)/(norm_d + eps);[value,idx] = max(Dtemp(fixed_d,f));Dtemp(:,f) = Dtemp(:,f)*sign(value); lambda(4,f,i) = sign(value)*(norm_d + eps);
        end
        if ~isempty(X2)
            norm_e = norm(Etemp(fixed_e,f),2);Etemp(:,f) = Etemp(:,f)/(norm_e + eps); lambda(5,f,i) = (norm_e + eps);
        end
        if ~isempty(X3)
            norm_g = norm(Gtemp(fixed_g,f),2);Gtemp(:,f) = Gtemp(:,f)/(norm_g + eps); lambda(6,f,i) = (norm_g + eps);
        end
       
    end
    
% %     paralleltime(i) = paralleltime(i) + toc;%add to parallelizable time

% %     tic;%start timing the serial part
    
    %Do the merge
    if i ==1
        A = Atemp; B = Btemp; C = Ctemp;
        
        if ~isempty(X1)
            D = Dtemp;
        end
        if ~isempty(X2)
            E = Etemp;
        end
        if ~isempty(X3)
            G = Gtemp;
        end
        
        
    else
        valA=zeros(F,1);valB=zeros(F,1);valC=zeros(F,1);
        valD=zeros(F,1);valE=zeros(F,1);valG=zeros(F,1);
        
        usedA = [];usedB = []; usedC=[]; usedD=[];usedE=[];usedG=[];%here we keep track of the columns we've already used
        for f1 = 1:F
            for f2 = 1:F
                valA(f2) = (Atemp(fixed_i,f1))'* (A(fixed_i,f2));
                valB(f2) = (Btemp(fixed_j,f1))'* (B(fixed_j,f2));
                valC(f2) = (Ctemp(fixed_k,f1))'* (C(fixed_k,f2));
                
                if ~isempty(X1)
                    valD(f2) = (Dtemp(fixed_d,f1))'* (D(fixed_d,f2));
                end
                if ~isempty(X2)
                    valE(f2) = (Etemp(fixed_e,f1))'* (E(fixed_e,f2));
                end
                if ~isempty(X3)
                    valG(f2) = (Gtemp(fixed_g,f1))'* (G(fixed_g,f2));
                end
               
            end
            
%             valA = abs(valA);
            [junk idx] = sort(valA,'descend');            
            ii = 1;while (find(idx(ii) == usedA))ii = ii +1; end;idx = idx(ii);usedA = [usedA idx];%use the highest non-already used matching column
            
            mask2 = find(A(:,idx)==0);
            A(mask2,idx) = A(mask2,idx) + Atemp(mask2,f1);%Update ONLY the zero values
            allCorrespondences(f1) = idx;
            
%             [junk idx] = sort(valB,'descend');            
%             ii = 1;while (find(idx(ii) == usedB))ii = ii +1; end;idx = idx(ii);usedB = [usedB idx];
            mask2 = find(B(:,idx)==0);
            B(mask2,idx) = B(mask2,idx) + Btemp(mask2,f1);%Update ONLY the zero values

            
%             [junk idx] = sort(valC,'descend');            
%             ii = 1;while (find(idx(ii) == usedC))ii = ii +1; end;idx = idx(ii);usedC = [usedC idx];            
            mask2 = find(C(:,idx)==0);
            C(mask2,idx) = C(mask2,idx) + Ctemp(mask2,f1);%Update ONLY the zero values

            
            
            
             if ~isempty(X1)
%                 [junk idx] = sort(valD,'descend');            
%                 ii = 1;while (find(idx(ii) == usedD))ii = ii +1; end;idx = idx(ii);usedD = [usedD idx];
                mask2 = find(D(:,idx)==0);
                D(mask2,idx) = D(mask2,idx) + Dtemp(mask2,f1);%Update ONLY the zero values
                
             end
             if ~isempty(X2)              
%                 [junk idx] = sort(valE,'descend');            
%                 ii = 1;while (find(idx(ii) == usedE))ii = ii +1; end;idx = idx(ii);usedE = [usedE idx];
                mask2 = find(E(:,idx)==0);
                E(mask2,idx) = E(mask2,idx) + Etemp(mask2,f1);%Update ONLY the zero values
                
                
             end  
             if ~isempty(X3)
                
%                 [junk idx] = sort(valG,'descend');            
%                 ii = 1;while (find(idx(ii) == usedG))ii = ii +1; end;idx = idx(ii);usedG = [usedG idx];
                mask2 = find(G(:,idx)==0);
                G(mask2,idx) = G(mask2,idx) + Gtemp(mask2,f1);%Update ONLY the zero values
            end
            
            
        end
        if i > 1
            lambda(:,:,i) = lambda(:,allCorrespondences,i);
        end
% %     serialtime = serialtime + toc;
    end
    

     %calculate final cost
    
%     tic; 
    cost = 0;
    if ~isempty(X1)
        cost = cost + norm(X1 - A*diag(lambda(1,:,i).*lambda(4,:,i))*D','fro')^2;
    end
    if ~isempty(X2)
        cost = cost + norm(X2 - B*diag(lambda(2,:,i).*lambda(5,:,i))*E','fro')^2;
    end
    if ~isempty(X3)
        cost = cost + norm(X3 - C*diag(lambda(3,:,i).*lambda(6,:,i))*G','fro')^2;
    end
    xx =  khatrirao(A,B)*diag((lambda(1,:,i)).*(lambda(2,:,i)).*(lambda(3,:,i)) )*C';

    cost = cost + norm(UC - xx ,'fro')^2;
    allcost(i) = cost;
%     serialtime = serialtime + toc;
end

    if times > 1
        if ~isempty(diff(allcost)./allcost(1:end-1) > 10^-2)
            errs = diff(allcost)./allcost(1:end-1);errs = sum(errs>10^-2);
            warning(sprintf('%d out of %d repetitions not monotonically decreasing cost',errs,times));
        end
    end

lambda = mean(lambda,3);
sigma = lambda(4,:);
A = A*diag(lambda(1,:)) ;
B = B*diag(lambda(2,:));
C = C*diag(lambda(3,:));
% lambda = prod(lambda(1:3,:),1); 
lambda = ones(F,1);

% A(isnan(A)) = 0;B(isnan(B)) = 0;C(isnan(C)) = 0;
A(abs(A)<10^-8) = 0;
B(abs(B)<10^-8) = 0;
C(abs(C)<10^-8) = 0;


                
 if ~isempty(X1)
%     D = sparse(D(:,:)*diag(lambda(4,:)));
    D(abs(D)<10^-8) = 0;
    D(isnan(D)) = 0;
    outmatrices{1} = sparse(D);
 else
     outmatrices{1} = [];
 end
 if ~isempty(X2)
%     E = sparse(E(:,:)*diag(lambda(5,:)));
    E(abs(E)<10^-8) = 0;
    E(isnan(E)) = 0;
    outmatrices{2} = sparse(E);
 else
     outmatrices{2} = [];
 end
 if ~isempty(X3)
%     G = sparse(G(:,:)*diag(lambda(6,:)));
    G(abs(G)<10^-8) = 0;
    G(isnan(G)) = 0;
    outmatrices{3} = sparse(G);
 else
     outmatrices{3} = [];
 end




end

%Vagelis Papalexakis, 2012
%School of Computer Science, Carnegie Mellon University
%Core sampling function for the ParCube algorithm, for memory resident
%tensors.
function [Xs X1s X2s X3s idx_i idx_j idx_k idx_d idx_e idx_g Xma Xmb Xmc] = coupled_parCube_core(X,X1,X2,X3,sample_factor,fixed_set)

X1s = []; X2s = []; X3s = [];

if numel(sample_factor)>1
    s1 = sample_factor(1); 
    s2 = sample_factor(2);
    s3 = sample_factor(3);
else
    s1 = sample_factor; 
    s2 = sample_factor;
    s3 = sample_factor;
end

if nargin == 6
   fixed_i = fixed_set{1};
   fixed_j = fixed_set{2};
   fixed_k = fixed_set{3};
   fixed_d = fixed_set{4};
   fixed_e = fixed_set{5};
   fixed_g = fixed_set{6};
else
    fixed_i = [];
    fixed_j = [];
    fixed_k = [];
    fixed_d = [];
    fixed_e = [];
    fixed_g = [];
end

% fixed_i
% pause

negative_values = find(X<0);
if ~isempty(negative_values)
    X_abs = X;
%     ii = find(X<0);
    X_abs(negative_values) = - X_abs(negative_values);
%     X_abs = double(X);
%     X_abs = abs(X_abs);
%     X_abs = sptensor(X_abs);
else
   X_abs = X; 
end
s = size(X); I = s(1); J = s(2); K = s(3);

if ~isempty(X1)
    Xma = collapse(X_abs,[2 3]) + sum(abs(X1),2);
else
    Xma = collapse(X_abs,[2 3]);
end
if ~isempty(X2)
    Xmb = collapse(X_abs,[1 3]) + sum(abs(X2),2);
else
     Xmb = collapse(X_abs,[1 3]);  
end
if ~isempty(X3)
    Xmc = collapse(X_abs,[1 2]) + sum(abs(X3),2);
else
    Xmc = collapse(X_abs,[1 2]); 
end

idx_i = find(Xma~=0);
idx_i = randsample_weighted(idx_i,I/s1,full(Xma(idx_i)));
if ~isempty(fixed_i)
    idx_i = union(idx_i,fixed_i);
end
idx_i = sort(idx_i);


idx_j = find(Xmb~=0);
idx_j = randsample_weighted(idx_j,J/s2,full(Xmb(idx_j)));
if ~isempty(fixed_j)
    idx_j = union(idx_j,fixed_j);
end
idx_j = sort(idx_j);

idx_k = find(Xmc~=0);
idx_k = randsample_weighted(idx_k,K/s3,full(Xmc(idx_k)));
if ~isempty(fixed_k)
    idx_k = union(idx_k,fixed_k);
end
idx_k = sort(idx_k);

X1d = sum(abs(X1),1);
X2e = sum(abs(X2),1);
X3g = sum(abs(X3),1);

idx_d = find(X1d~=0);
idx_d = randsample_weighted(idx_d,size(X1,2)/s1,full(X1d(idx_d)));
if ~isempty(fixed_d)
    idx_d = union(idx_d,fixed_d);
end
idx_d = sort(idx_d);

idx_e = find(X2e~=0);
idx_e = randsample_weighted(idx_e,size(X2,2)/s2,full(X2e(idx_e)));
if ~isempty(fixed_e)
    idx_e = union(idx_e,fixed_e);
end
idx_e = sort(idx_e);

idx_g = find(X3g~=0);
idx_g = randsample_weighted(idx_g,size(X3,2)/s3,full(X3g(idx_g)));
if ~isempty(fixed_g)
    idx_g = union(idx_g,fixed_g);
end
idx_g = sort(idx_g);


if s1 ==1
    idx_i = 1:I;
end
if s2 == 1
    idx_j = 1:J;
end
if s3 == 1
    idx_k = 1:K;
end


% X = double(X);
Xs = sptensor(X(idx_i,idx_j,idx_k));
if ~isempty(X1)
    X1s = X1(idx_i,idx_d);
end
if ~isempty(X2)
    X2s = X2(idx_j,idx_e);
end
if ~isempty(X3)
    X3s = X3(idx_k,idx_g);
end

end