function [A B C outmatrices lambda sigma] = TurboSMT_als_parallel(X,matrices,F,sample_factor,times,nonneg,p)
%Vagelis Papalexakis, 2013-2014
%School of Computer Science, Carnegie Mellon University
%Implementation of the multi-core version of Turbo-SMT Algorithm using the
%CMTF-ALS as core algorithm


if nargin == 5
    nonneg = 0;
    p = 0.3;
end
if nargin == 6
    p=0.3;
end
mypath = pwd;


s = size(X); I = s(1); J = s(2); K = s(3);

%handle the additional matrices
mat = {};
for i = 1:length(matrices)
   mat{i} = matrices{i}; 
end
for i = length(matrices)+1 : 3
    mat{i} = [];
end
matrices = mat;

X1 = matrices{1}; X2 = matrices{2}; X3 = matrices{3};

if ~isempty(X1)
    if size(X1,1)~=I
       error('X1 must have I rows'); 
    end
end

if ~isempty(X2)
    if size(X2,1)~=J
       error('X2 must have J rows'); 
    end
end

if ~isempty(X3)
    if size(X3,1)~=K
       error('X3 must have K rows'); 
    end
end
% tic

A = sparse(I,F); B = sparse(J,F);C = sparse(K,F);

[Xs X1s X2s X3s idx_i idx_j idx_k idx_d idx_e idx_g Xma Xmb Xmc] = coupled_parCube_core(X,X1,X2,X3,sample_factor);

fixed_i = idx_i(1:ceil(p*length(idx_i)));
fixed_j = idx_j(1:ceil(p*length(idx_j)));
fixed_k = idx_k(1:ceil(p*length(idx_k)));
fixed_d = idx_d(1:ceil(p*length(idx_d)));
fixed_e = idx_e(1:ceil(p*length(idx_e)));
fixed_g = idx_g(1:ceil(p*length(idx_g)));

fixed{1} = fixed_i; fixed{2} = fixed_j; fixed{3}=fixed_k;
fixed{4} = fixed_d; fixed{5} = fixed_e; fixed{6} = fixed_g;
% serialtime = serialtime + toc;

system('mkdir tmp/');
% tic %start timing parallelizable portion
parfor ipar = 1:times

    lambdaPar = zeros(6, F);%hold scaling factor for each column of all matrices
    
    
    [Xs X1s X2s X3s idx_i idx_j idx_k idx_d idx_e idx_g Xma Xmb Xmc] = coupled_parCube_core(X,X1,X2,X3,sample_factor,fixed);     
    mats = cell(3);
    mats{1} = []; mats{2} = []; mats{3} = [];
    if ~isempty(X1)
        mats{1} = X1s;
    end
    if ~isempty(X2)
        mats{2} = X2s;
    end
    if ~isempty(X3)
        mats{3} = X3s;
    end
    if nnz(Xs)>0
        if nonneg
            [As Bs Cs OUTMs] = NNCMTF_dense(Xs,mats,F);
        else
            [As Bs Cs OUTMs] = fastCMTF_dense(Xs,mats,F);
        end
    else
       As = 0; Bs = 0; Cs = 0; OUTMs = {};
    end
    
    

    AtempPar = sparse(I,F); BtempPar = sparse(J,F);CtempPar = sparse(K,F);


    AtempPar(idx_i,:) = As; BtempPar(idx_j,:) = Bs; CtempPar(idx_k,:) = Cs;
    
    
    if ~isempty(X1)
        DtempPar = sparse(size(X1,2),F); 
        DtempPar(idx_d,:) = OUTMs{1};
    else
        DtempPar = [];
    end
    if ~isempty(X2)
        EtempPar = sparse(size(X2,2),F);
        EtempPar(idx_e,:) = OUTMs{2};
    else
        EtempPar = [];
    end
    if ~isempty(X3)
        GtempPar = sparse(size(X3,2),F);    
        GtempPar(idx_g,:) = OUTMs{3};
    else
        GtempPar = [];
    end
   
    
     %now, normalize using the scaling of the common part of every column
    for f = 1:F
       norm_a = norm(AtempPar(fixed_i,f),2);AtempPar(:,f) = AtempPar(:,f)/(norm_a + eps); lambdaPar(1,f) = (norm_a + eps);
       norm_b = norm(BtempPar(fixed_j,f),2);BtempPar(:,f) = BtempPar(:,f)/(norm_b + eps); lambdaPar(2,f) = (norm_b + eps);
       norm_c = norm(CtempPar(fixed_k,f),2);CtempPar(:,f) = CtempPar(:,f)/(norm_c + eps); lambdaPar(3,f) = (norm_c + eps);
        
       %consider cases for all possible numbers of out matrices
        if ~isempty(X1)
           norm_d = norm(DtempPar(fixed_d,f),2);DtempPar(:,f) = DtempPar(:,f)/(norm_d + eps); lambdaPar(4,f) = (norm_d + eps);
        end
        if ~isempty(X2)
            norm_e = norm(EtempPar(fixed_e,f),2);EtempPar(:,f) = EtempPar(:,f)/(norm_e + eps); lambdaPar(5,f) = (norm_e + eps);
        end
        if ~isempty(X3)
            norm_g = norm(GtempPar(fixed_g,f),2);GtempPar(:,f) = GtempPar(:,f)/(norm_g + eps); lambdaPar(6,f) = (norm_g + eps);
        end
       
    end
    
    
    [ii jj kk] = find(AtempPar);data_dump = [ii jj kk];dlmwrite(sprintf('tmp/Atemp%d',ipar),data_dump,'\t');
    [ii jj kk] = find(BtempPar);data_dump = [ii jj kk];dlmwrite(sprintf('tmp/Btemp%d',ipar),data_dump,'\t');
    [ii jj kk] = find(CtempPar);data_dump = [ii jj kk];dlmwrite(sprintf('tmp/Ctemp%d',ipar),data_dump,'\t');
    [ii jj kk] = find(DtempPar);data_dump = [ii jj kk];dlmwrite(sprintf('tmp/Dtemp%d',ipar),data_dump,'\t');
    [ii jj kk] = find(lambdaPar);data_dump = [ii jj kk];dlmwrite(sprintf('tmp/lambda%d',ipar),data_dump,'\t');


%save as AtempPar.mat and then 
%Atemp = load('AtempPar.mat');Atemp = Atemp.AtempPar;
end  
    

%     paralleltime = toc;%add to parallelizable time

    
 lambda = zeros(4, F,times);%hold scaling factor for each column of all matrices, for all repetitions
 
%     tic;%start timing the serial part
for i =1:times  
   
        
    data_dump = dlmread(sprintf('tmp/Atemp%d',i),'\t');Atemp = sparse(data_dump(:,1),data_dump(:,2),data_dump(:,3),size(X,1),F);
    data_dump = dlmread(sprintf('tmp/Btemp%d',i),'\t');Btemp = sparse(data_dump(:,1),data_dump(:,2),data_dump(:,3),size(X,2),F);
    data_dump = dlmread(sprintf('tmp/Ctemp%d',i),'\t');Ctemp = sparse(data_dump(:,1),data_dump(:,2),data_dump(:,3),size(X,3),F);
    data_dump = dlmread(sprintf('tmp/Dtemp%d',i),'\t');Dtemp = sparse(data_dump(:,1),data_dump(:,2),data_dump(:,3),size(X1,2),F);
    data_dump = dlmread(sprintf('tmp/lambda%d',i),'\t');lambdaTemp = spconvert(data_dump);
    lambda(:,:,i) = lambdaTemp;

    
    
    allCorrespondences = zeros(F,1);


    %Do the merge
    if i ==1
        A = Atemp; B = Btemp; C = Ctemp;
        
        if ~isempty(X1)
            D = Dtemp;
        end
        if ~isempty(X2)
            E = Etemp;
        end
        if ~isempty(X3)
            G = Gtemp;
        end
        
        
    else
        valA=zeros(F,1);valB=zeros(F,1);valC=zeros(F,1);
        valD=zeros(F,1);valE=zeros(F,1);valG=zeros(F,1);
        
        
        usedA = [];usedB = []; usedC=[]; usedD=[];usedE=[];usedG=[];%here we keep track of the columns we've already used
        for f1 = 1:F
            for f2 = 1:F
                valA(f2) = (Atemp(fixed_i,f1))'* (A(fixed_i,f2));
                size(Btemp)
                size(B)
                
                Btemp(fixed_j,f1)
                B(fixed_j,f2)
                valB(f2) = (Btemp(fixed_j,f1))'* (B(fixed_j,f2));
                valC(f2) = (Ctemp(fixed_k,f1))'* (C(fixed_k,f2));
                
                if ~isempty(X1)
                    valD(f2) = (Dtemp(fixed_d,f1))'* (D(fixed_d,f2));
                end
                if ~isempty(X2)
                    valE(f2) = (Etemp(fixed_e,f1))'* (E(fixed_e,f2));
                end
                if ~isempty(X3)
                    valG(f2) = (Gtemp(fixed_g,f1))'* (G(fixed_g,f2));
                end
                
            end
            
            
            [junk idx] = sort(valA,'descend');            
            ii = 1;while (find(idx(ii) == usedA))ii = ii +1; end;idx = idx(ii);usedA = [usedA idx];%use the highest non-already used matching column
            mask2 = find(A(:,idx)==0);
            A(mask2,idx) = A(mask2,idx) + Atemp(mask2,f1);%Update ONLY the zero values
            
            allCorrespondences(f1) = idx;
% %             [junk idx] = sort(valB,'descend');            
% %             ii = 1;while (find(idx(ii) == usedB))ii = ii +1; end;idx = idx(ii);usedB = [usedB idx];
            mask2 = find(B(:,idx)==0);
            B(mask2,idx) = B(mask2,idx) + Btemp(mask2,f1);%Update ONLY the zero values

            
% %             [junk idx] = sort(valC,'descend');            
% %             ii = 1;while (find(idx(ii) == usedC))ii = ii +1; end;idx = idx(ii);usedC = [usedC idx];            
            mask2 = find(C(:,idx)==0);
            C(mask2,idx) = C(mask2,idx) + Ctemp(mask2,f1);%Update ONLY the zero values

            
            
            
             if ~isempty(X1)
% %                 [junk idx] = sort(valD,'descend');            
% %                 ii = 1;while (find(idx(ii) == usedD))ii = ii +1; end;idx = idx(ii);usedD = [usedD idx];
                mask2 = find(D(:,idx)==0);
                D(mask2,idx) = D(mask2,idx) + Dtemp(mask2,f1);%Update ONLY the zero values
                
             end
             if ~isempty(X2)              
% %                 [junk idx] = sort(valE,'descend');            
% %                 ii = 1;while (find(idx(ii) == usedE))ii = ii +1; end;idx = idx(ii);usedE = [usedE idx];
                mask2 = find(E(:,idx)==0);
                E(mask2,idx) = E(mask2,idx) + Etemp(mask2,f1);%Update ONLY the zero values
                
                
             end  
             if ~isempty(X3)
                
% %                 [junk idx] = sort(valG,'descend');            
% %                 ii = 1;while (find(idx(ii) == usedG))ii = ii +1; end;idx = idx(ii);usedG = [usedG idx];
                mask2 = find(G(:,idx)==0);
                G(mask2,idx) = G(mask2,idx) + Gtemp(mask2,f1);%Update ONLY the zero values
            end
            
            
        end
    %reorder lambda
    lambda(:,:,i) = lambda(:,allCorrespondences,i);
    end

end
% serialtime = serialtime + toc;


lambda = mean(lambda,3);
sigma = lambda(4,:);
lambda = prod(lambda(1:3,:),1); 


% A(isnan(A)) = 0;B(isnan(B)) = 0;C(isnan(C)) = 0;
A(abs(A)<10^-8) = 0;
B(abs(B)<10^-8) = 0;
C(abs(C)<10^-8) = 0;

                
 if ~isempty(X1)
%     D = sparse(D(:,:)*diag(lambda(4,:)));
    D(abs(D)<10^-8) = 0;
    D(isnan(D)) = 0;
    outmatrices{1} = sparse(D);
 else
     outmatrices{1} = [];
 end
 if ~isempty(X2)
%     E = sparse(E(:,:)*diag(lambda(5,:)));
    E(abs(E)<10^-8) = 0;
    E(isnan(E)) = 0;
    outmatrices{2} = sparse(E);
 else
     outmatrices{2} = [];
 end
 if ~isempty(X3)
%     G = sparse(G(:,:)*diag(lambda(6,:)));
    G(abs(G)<10^-8) = 0;
    G(isnan(G)) = 0;
    outmatrices{3} = sparse(G);
 else
     outmatrices{3} = [];
 end




end

%Vagelis Papalexakis, 2012
%School of Computer Science, Carnegie Mellon University
%Core sampling function for the ParCube algorithm, for memory resident
%tensors.
function [Xs X1s X2s X3s idx_i idx_j idx_k idx_d idx_e idx_g Xma Xmb Xmc] = coupled_parCube_core(X,X1,X2,X3,sample_factor,fixed_set)

X1s = []; X2s = []; X3s = [];

if numel(sample_factor)>1
    s1 = sample_factor(1); 
    s2 = sample_factor(2);
    s3 = sample_factor(3);
else
    s1 = sample_factor; 
    s2 = sample_factor;
    s3 = sample_factor;
end

if nargin == 6
   fixed_i = fixed_set{1};
   fixed_j = fixed_set{2};
   fixed_k = fixed_set{3};
   fixed_d = fixed_set{4};
   fixed_e = fixed_set{5};
   fixed_g = fixed_set{6};
else
    fixed_i = [];
    fixed_j = [];
    fixed_k = [];
    fixed_d = [];
    fixed_e = [];
    fixed_g = [];
end

% fixed_i
% pause

negative_values = find(X<0);
if ~isempty(negative_values)
    X_abs = X;
%     ii = find(X<0);
    X_abs(negative_values) = - X_abs(negative_values); 
%     X_abs = double(X);
%     X_abs = abs(X_abs);
%     X_abs = sptensor(X_abs);
else
   X_abs = X; 
end
s = size(X); I = s(1); J = s(2); K = s(3);

if ~isempty(X1)
    Xma = collapse(X_abs,[2 3]) + sum(abs(X1),2);
else
    Xma = collapse(X_abs,[2 3]);
end
if ~isempty(X2)
    Xmb = collapse(X_abs,[1 3]) + sum(abs(X2),2);
else
     Xmb = collapse(X_abs,[1 3]);  
end
if ~isempty(X3)
    Xmc = collapse(X_abs,[1 2]) + sum(abs(X3),2);
else
    Xmc = collapse(X_abs,[1 2]); 
end

idx_i = find(Xma~=0);
idx_i = randsample_weighted(idx_i,I/s1,double(Xma(idx_i)));
if ~isempty(fixed_i)
    idx_i = union(idx_i,fixed_i);
end
idx_i = sort(idx_i);


idx_j = find(Xmb~=0);
idx_j = randsample_weighted(idx_j,J/s2,double(Xmb(idx_j)));
if ~isempty(fixed_j)
    idx_j = union(idx_j,fixed_j);
end
idx_j = sort(idx_j);

idx_k = find(Xmc~=0);
idx_k = randsample_weighted(idx_k,K/s3,double(Xmc(idx_k)));
if ~isempty(fixed_k)
    idx_k = union(idx_k,fixed_k);
end
idx_k = sort(idx_k);

X1d = sum(abs(X1),1);
X2e = sum(abs(X2),1);
X3g = sum(abs(X3),1);

idx_d = find(X1d~=0);
idx_d = randsample_weighted(idx_d,size(X1,2)/s1,full(X1d(idx_d)));
if ~isempty(fixed_d)
    idx_d = union(idx_d,fixed_d);
end
idx_d = sort(idx_d);

idx_e = find(X2e~=0);
idx_e = randsample_weighted(idx_e,size(X2,2)/s2,full(X2e(idx_e)));
if ~isempty(fixed_e)
    idx_e = union(idx_e,fixed_e);
end
idx_e = sort(idx_e);

idx_g = find(X3g~=0);
idx_g = randsample_weighted(idx_g,size(X3,2)/s3,full(X3g(idx_g)));
if ~isempty(fixed_g)
    idx_g = union(idx_g,fixed_g);
end
idx_g = sort(idx_g);


if s1 ==1
    idx_i = 1:I;
end
if s2 == 1
    idx_j = 1:J;
end
if s3 == 1
    idx_k = 1:K;
end


% X = double(X);
Xs = tensor(X(idx_i,idx_j,idx_k));
if ~isempty(X1)
    X1s = X1(idx_i,idx_d);
end
if ~isempty(X2)
    X2s = X2(idx_j,idx_e);
end
if ~isempty(X3)
    X3s = X3(idx_k,idx_g);
end

end

function parsave(fname, x)
save(fname, 'x')
end
